// partijcombinaties bepalen
// Evert Mouw, 2010-06-10
// 2010-06-11: nu met uitsluitingen
// 2010-06-12: nu met normalisatie

#include <iostream>
#include <string>
#include <cmath>
#include <vector>
using namespace std;

// constants
const int AANTALPARTIJEN = 10;
const int MAXAANTALPARTIJEN = 10;
const int MINIMUM = 76;

// data
struct s_partij {
	string naam;
	int zetels;
	int x;
	int y;
};
s_partij partij[AANTALPARTIJEN];

void gegevens() {
	partij[0].naam = "vvd";
	partij[0].zetels = 31;
	partij[0].x = 387;
	partij[0].y = 319;

	partij[1].naam = "pvda";
	partij[1].zetels = 30;
	partij[1].x = 169;
	partij[1].y = 212;

	partij[2].naam = "pvv";
	partij[2].zetels = 24;
	partij[2].x = 259;
	partij[2].y = 364;

	partij[3].naam = "cda";
	partij[3].zetels = 21;
	partij[3].x = 333;
	partij[3].y = 372;

	partij[4].naam = "sp";
	partij[4].zetels = 15;
	partij[4].x = 82;
	partij[4].y = 196;

	partij[5].naam = "gl";
	partij[5].zetels = 10;
	partij[5].x = 175;
	partij[5].y = 166;

	partij[6].naam = "d66";
	partij[6].zetels = 10;
	partij[6].x = 280;
	partij[6].y = 165;

	partij[7].naam = "cu";
	partij[7].zetels = 5;
	partij[7].x = 200;
	partij[7].y = 324;

	partij[8].naam = "pvdd";
	partij[8].zetels = 2;
	partij[8].x = 175;
	partij[8].y = 203;

	partij[9].naam = "sgp";
	partij[9].zetels = 2;
	partij[9].x = 292;
	partij[9].y = 425;
}

// afstanden

int afstand[AANTALPARTIJEN][AANTALPARTIJEN];

int berekenafstand (int a, int b) {
	// a en b komen overeen met partij[a] en partij[b]
	int x = partij[b].x - partij[a].x;
	int y = partij[b].y - partij[a].y;
	return sqrt ( pow(x,2) + pow(y,2) );
}

int berekenalleafstanden() {
	int h;
	for (int i=0; i<AANTALPARTIJEN; i++) {
		for (int j=i+1; j<AANTALPARTIJEN; j++) {
			if (i==j) afstand[i][j]=0;
			else {
				h = berekenafstand(i,j);
				afstand[i][j] = h;
				afstand[j][i] = h;
			}
		}
	}
}

void toonalleafstanden() {
	for (int i=0; i<AANTALPARTIJEN; i++) {
		for (int j=i+1; j<AANTALPARTIJEN; j++) {
			cout << "afstand " << partij[i].naam << " - " << partij[j].naam << " is " << afstand[i][j] << endl;
		}
	}
}

void normaliseerafstanden() {
	// deze functie moet uitgevoerd worden VOORDAT andere functies uitgevoerd worden
	// die ook gebruik maken van gegevens()  en dus van partij[]
	
	// initialisatie is afhankelijk van coordinaatsysteem
	// beginwaarden moeten in de oorsprong liggen
	int xmax = 258, xmin = 258, ymax = 273, ymin = 273;
	
	// eerst de grootste en kleinste waarden voor x en y vinden
	for (int i=0; i<AANTALPARTIJEN; i++) {
		if ( partij[i].x > xmax ) xmax = partij[i].x;
		if ( partij[i].x < xmin ) xmin = partij[i].x;
		if ( partij[i].y > ymax ) ymax = partij[i].y;
		if ( partij[i].y < ymin ) ymin = partij[i].y;
	}
	
	// TESTING
	//cout << "xmin = " << xmin << endl;
	//cout << "xmax = " << xmax << endl;
	//cout << "ymin = " << ymin << endl;
	//cout << "ymax = " << ymax << endl;
	
	// met ... vermenigvuldigen om te normaliseren: grootste afstand is 100
	float xfactor = 100 / float(xmax - xmin);
	float yfactor = 100 / float(ymax - ymin);
	
	// TESTING
	//cout << "xfactor = " << xfactor << endl;
	//cout << "yfactor = " << yfactor << endl;
	
	// de afstanden nu daadwerkelijk normaliseren
	for (int i=0; i<AANTALPARTIJEN; i++) {
		partij[i].x = (float)partij[i].x * xfactor;
		partij[i].y = (float)partij[i].y * yfactor;
	}
	
	// TESTING
	cout << "normalisatie / ijking:" << endl;
	cout << "  afstand X voor D66 - SGP is " << partij[9].x - partij[6].x << endl;
	cout << "  afstand Y voor D66 - SGP is " << partij[9].y - partij[6].y << endl;
	cout << "  afstand X voor SP - VVD is " << partij[0].x - partij[4].x << endl;
	cout << "  afstand Y voor SP - VVD is " << partij[0].y - partij[4].y << endl;
}

// uitsluitingen

int uitsluitingen() {
	// sommige partijen willen echt niet met elkaar samenwerken
	// in dat geval verdubbelen we de onderlinge afstand
	cout << "Sommige partijen willen echt niet met elkaar samenwerken." << endl;
	cout << "Die uitsluitingscombinaties krijgen een verdubbelde onderlinge afstand." << endl;
	cout << "Uitsluiten van samenwerking:" << endl;

	// pvda - pvv
	cout << "* pvda - pvv" << endl;
	afstand[1][2] *= 2;
	afstand[2][1] *= 2;

	// pvv - gl
	cout << "* pvv - gl" << endl;
	afstand[2][5] *= 2;
	afstand[5][2] *= 2;

	// pvv - d66
	cout << "* pvv - d66" << endl;
	afstand[2][6] *= 2;
	afstand[6][2] *= 2;

}

class c_combinatie {
	private:
	vector<int> partijen;
	int som_zetels;
	int som_afstanden;
	void update_som_zetels() {
		som_zetels = 0;
		for (int i=0; i<partijen.size(); i++) {
			som_zetels += partij[partijen[i]].zetels;
		}
	}
	void update_som_afstanden() {
		som_afstanden = 0;
		if ( partijen.size() >= 2 ) {
			for (int i=0; i<partijen.size(); i++) {
				for (int j=i+1; j<partijen.size(); j++) {
					som_afstanden += afstand[partijen[i]][partijen[j]];
				}
			}
		}
	}
	public:
	c_combinatie() {
		som_zetels = 0;
		som_afstanden = 0;
	}
	void voegtoe(int partij) {
		partijen.push_back (partij);
		update_som_zetels();
		update_som_afstanden();
	}
	void tonen() {
		for (int i=0; i<partijen.size(); i++) {
			cout << partij[partijen[i]].naam << " ";
		}
		cout << " | som_zetels: " << som_zetels;
		cout << " | som_afstanden: " << som_afstanden;
		cout << endl;
	}
	int grootte() {
		return partijen.size();
	}
	int zetels() {
		return som_zetels;
	}
	int afstanden() {
		return som_afstanden;
	}
	int laatstepartij() {
		if ( partijen.size() ) return partijen.back();
		else return -1;
	}
	int geefpartij(int nummer) {
		if ( nummer < partijen.size() ) return partijen.at(nummer);
		else return -1;
	}
};

void maakcombi (vector<c_combinatie*> * combinaties, int maxgrootte, c_combinatie * combi) {
	// unieke combinaties maken dmv recursie

	// ben ik de eerste?
	if (combi==NULL) combi = new c_combinatie;

	// einde test
	if ( combi->grootte() >= maxgrootte ) {
		combinaties->push_back(combi);
		return;
	}

	// voor unieke combinaties: nieuwe toevoeging > laatste toevoeging
	int test = combi->laatstepartij();
	test++;
	for (int i=test; i<AANTALPARTIJEN; i++) {
		// nieuw object maken
		c_combinatie * nieuwecombi = new c_combinatie;
		for (int n=0; n<combi->grootte(); n++) {
			nieuwecombi->voegtoe( combi->geefpartij(n) );
		}
		nieuwecombi->voegtoe(i);
		maakcombi(combinaties, maxgrootte, nieuwecombi);
	}
	delete combi;
}

int main() {
	int n, k;
	gegevens();
	normaliseerafstanden();
	berekenalleafstanden();
	uitsluitingen();
	toonalleafstanden();

	vector<c_combinatie*> * combinaties = new vector<c_combinatie*>;
	c_combinatie combi;

	// maak alle combinaties, zonder te kijken naar het zetelaantal
	for (int g=1; g<AANTALPARTIJEN+1; g++) { // grootte van de combi
		maakcombi (combinaties, g, NULL);
	}

	// verwijder alle combinaties met te weinig zetels
	n = 0;
	while ( n < combinaties->size() ) {
		if ( combinaties->at(n)->zetels() < MINIMUM ) {
			combinaties->erase( combinaties->begin() + n );
		}
		else n++;
	}

	// sorteer nu op afstand (simpele bubble sort)
	c_combinatie * c = NULL;
	for (int i=0; i<combinaties->size(); i++) {
		for (int j=i+1; j<combinaties->size(); j++) {
			if ( combinaties->at(i)->afstanden() > combinaties->at(j)->afstanden() ) {
				c = combinaties->at(j);
				combinaties->at(j) = combinaties->at(i);
				combinaties->at(i) = c;
			}
		}
	}

	// laat alle combinaties zien
	for (int i=0; i<combinaties->size(); i++) {
		combinaties->at(i)->tonen();
	}

	return 0;
}
